/**
 * SCTP implementation according to RFC 2960.
 * Copyright (C) 2000 by Siemens AG, Munich, Germany.
 *
 * Realized in co-operation between Siemens AG
 * and University of Essen, Institute of Computer Networking Technology.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * There are two mailinglists available at www.sctp.de which should be used for
 * any discussion related to this implementation.
 *
 * Contact: discussion@sctp.de
 *          Michael.Tuexen@icn.siemens.de
 *          ajung@exp-math.uni-essen.de
 *
 *
 *
 */
import java.io.*;
import java.net.*;

public class SockReaderThread extends Thread  {

    public static final byte ULP_REGISTER=1;
    public static final byte ULP_REGISTER_OK=2;
    public static final byte ULP_DATA_ARRIVE=7;
    public static final byte ULP_SEND_FAILURE=            8;
    public static final byte ULP_NETWORK_STATUS_CHANGE=   9;
    public static final byte ULP_COMM_UP=                 10;
    public static final byte ULP_COMM_LOST=               11;
    public static final byte ULP_COMM_ERR=                12;
    public static final byte ULP_RESTART_NOTIFICATION=    13;
    public static final byte ULP_SHUTDOWN_COMPLETE=       14;
    public static final byte ULP_INITIALIZE_OK=           15;
    public static final byte ULP_ASSOCIATE_OK=            16;
    public static final byte ULP_SHUTDOWN_OK=             17;
    public static final byte ULP_CLOSE_OK=                18;
    public static final byte ULP_SEND_OK=                 19;
    public static final byte ULP_SET_PRIMARY_OK=          20;
    public static final byte ULP_RECEIVE_OK=              21;
    public static final byte ULP_STATUS_OK=               22;
    public static final byte ULP_CHANGE_HB_OK=            23;
    public static final byte ULP_REQ_HB_OK=               24;
    public static final byte ULP_GET_SRTT_OK=             25;
    public static final byte ULP_INITIALIZE=28;
    public static final byte ULP_ASSOCIATE=29;
    public static final byte ULP_SHUTDOWN=30;
    public static final byte ULP_CLOSE=31;
    public static final byte ULP_SEND=32;
    public static final byte ULP_SET_PRIMARY=33;
    public static final byte ULP_RECEIVE=34;


    DatagramSocket ulpSocket;
    short localPort, peerPort;
    AssociationClass myAssoc;
    NotificationClass myNotifier;

    PrintWriter out;
    BufferedReader in;
    DataInputStream dis;
    ByteArrayInputStream bis;
    DataOutputStream dos;
    ByteArrayOutputStream bos;
	
	static String testString = "This is a test";

    public SockReaderThread(AssociationClass association,
                            NotificationClass notification,
                            short LocalPeerPort,
                            short LocalSrcPort,
                            DatagramSocket udpSocket) {
        // super(this);
        ulpSocket = udpSocket;
        localPort =  LocalSrcPort;
        peerPort =   LocalPeerPort;
        myAssoc =  association;
        myNotifier =  notification;
    }
		
	public void run() {
		byte[] rbuf = new byte[3000];
        int length;
        byte rtype = 0;
        int assocId = 0;
        int streamId = 0;
        int dataLen = 0;
  		DatagramPacket rpacket = null;
        while (true) {
            try {
        		rpacket = new DatagramPacket(rbuf, rbuf.length);
                ulpSocket.receive(rpacket);
                length = rpacket.getLength();
        		dis = new DataInputStream(new ByteArrayInputStream(rbuf));
	        	rtype = dis.readByte();
            } catch (Exception e) {
                System.err.println("Receive Packet : Caught Exception: " + e.getMessage());
            }

            switch(rtype) {
                case ULP_REGISTER_OK :
                    System.err.println("Received REGISTER_OK from Peer - should have been caught by main thread !");
                    System.exit(-1);
                    break;
                case ULP_DATA_ARRIVE :
                    // decode packet, and get message
                    try {
                        assocId = dis.readInt();
                        streamId = dis.readInt();
                        dataLen = dis.readInt();
                    } catch (Exception e) {
                        System.err.println("Caught Exception: " + e.getMessage());
                    }

                    System.out.print("Data Arrive Notification : Assoc=");
                    System.out.print(assocId);
                    System.out.print(" Stream ID=");
                    System.out.print(streamId);
                    System.out.print(" Length=");
                    System.out.println(dataLen);

                    System.out.println(doUlpReceive(rbuf,assocId, streamId, dataLen));
                    break;
                case ULP_SEND_FAILURE :
                    System.out.println("Received SEND_FAILURE !");
                    break;
                case ULP_NETWORK_STATUS_CHANGE :
                    System.out.println("Network Status Change - Parse Data not implemented yet !");
                    break;
                case ULP_COMM_UP :
                    try {
                        assocId = dis.readInt();
                        int numOfDestinations = dis.readInt();
                        short state = dis.readShort();
                        short numOfInStreams = dis.readShort();
                        short numOfOutStreams = dis.readShort();
                        System.out.print("Comm Up :  assocId=");
                        System.out.print(assocId);
                        System.out.print(" State=");
                        System.out.print(state);
                        System.out.print(" numOfDestinations=");
                        System.out.println(numOfDestinations);
                        System.out.print("  numOfInStreams=");
                        System.out.print( numOfInStreams);
                        System.out.print("  numOfOutStreams=");
                        System.out.println(  numOfOutStreams);

                        myAssoc.commUpNotification(assocId, state, numOfDestinations, numOfInStreams, numOfOutStreams);
                    } catch (Exception e) {
                        System.err.println("COMM UP : Caught Exception: " + e.getMessage());
                    }
                    break;
                case ULP_COMM_LOST :
                    System.out.println("COMMUNICATION LOST !");
                    break;
                case ULP_COMM_ERR :
                    System.out.println("COMMUNICATION ERROR !");
                    break;
                case ULP_RESTART_NOTIFICATION :
                    System.out.println("RESTART NOTIFICATION !");
                    break;
                case ULP_SHUTDOWN_COMPLETE :
                    System.out.println("SHUTDOWN COMPLETE -- Bye bye !!!!!!!!!!");
                    System.exit(0);
                    break;

                case ULP_INITIALIZE_OK :
                case ULP_ASSOCIATE_OK :
                case ULP_SHUTDOWN_OK  :
                case ULP_CLOSE_OK     :
                case ULP_SEND_OK      :
                case ULP_SET_PRIMARY_OK :
                case ULP_RECEIVE_OK     :
                case ULP_STATUS_OK      :
                case ULP_CHANGE_HB_OK   :
                case ULP_REQ_HB_OK      :
                case ULP_GET_SRTT_OK    :
    		    	System.out.print("Received message id = (");
                    System.out.print(rtype);
                    System.out.println(") from peer !");
                    myNotifier.put(rpacket);
                    break;
                default :
        			System.err.print("Received unknown message id  from Peer SCTP instance (");
                    System.err.print(rtype);
                    System.err.println(") !");
                    return;
            }
        }
	}

    private String doUlpReceive(byte[] recvBuf, int assocId, int streamId, int dataLen) {
        byte rtype = 0;
        byte [] buf = new byte[3000];
  		byte[] sbuf = null;
        int result;
        String s = null;

		bos = new ByteArrayOutputStream(100);
		dos = new DataOutputStream(bos);

        try {
            Integer sid = new Integer(streamId);
    		dos.writeByte(ULP_RECEIVE);
	    	dos.writeInt(myAssoc.getAssocId());  /* the assoc id */
		    dos.writeShort(sid.shortValue());  /* stream id  */
    		dos.writeInt(dataLen);

       		sbuf = bos.toByteArray();
        	DatagramPacket packet = new DatagramPacket(sbuf,sbuf.length,InetAddress.getByName("127.0.0.1"),peerPort);
            ulpSocket.send(packet);	
        } catch (Exception e) {
            System.err.println("i/o exception when sending user Data !"+ e.getMessage());
            System.exit(1);
        }
      	System.out.println("doUlpReceive : sending done !");

        try {
      		DatagramPacket rpacket = new DatagramPacket(recvBuf, recvBuf.length);
            ulpSocket.receive(rpacket);
		    DataInputStream dis = new DataInputStream(new ByteArrayInputStream(recvBuf));
    		rtype = dis.readByte();

    		if (rtype == ULP_RECEIVE_OK) {
                result = dis.readInt();
                s = new String(recvBuf, 5, dataLen);
             } else {
                System.err.print("Got unexpected type ");
                System.err.print(rtype);
                System.err.println(" instead of ULP_RECEIVE_OK !");
                s = "ERROR" ;
             }
        } catch (Exception e) {
            System.err.println("exception caught !"+ e.getMessage());
            System.exit(1);
        }

        return s;
    }

}

