#include <sys/param.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <sys/time.h>
#include <sys/signal.h>

#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <errno.h>

// #include <linux/icmp.h>
#ifndef ICMP_FILTER
	#define ICMP_FILTER                     1
#endif

typedef struct SCTP_COMMON_HEADER
{
    unsigned short src_port;
    unsigned short dest_port;
    unsigned int verification_tag;
    unsigned int checksum;
} SCTP_common_header;



#define ip_hl ihl
#define ip_v version
#define ip_tos tos
#define ip_len tot_len
#define ip_id id
#define ip_off frag_off
#define ip_ttl ttl
#define ip_p protocol
#define ip_sum check
#define ip_src saddr
#define ip_dst daddr

//  #define ICMP_MINLEN     28
#define MY_PACKETLEN 	1500
struct sockaddr whereto;      /* who to ping */
int datalen = MY_PACKETLEN;
int s;                        /* socket file descriptor */
u_char outpack[MY_PACKETLEN];    
static char *hostname;        
static int ident;       /* process id to identify our packets */


static void
usage(void)
{
      (void)fprintf(stderr, "usage: testraw  host-ipaddress\n");
      exit(2);
}

/*
 * pr_pack --
 * ALL readers of the ICMP socket get a copy of ALL ICMP packets
 * which arrive ('tis only fair).  Take care of that !
 * Taken from ping.c
 */
void
pr_pack(char *buf, int cc, struct sockaddr_in *from)
{     
	register struct icmphdr *icp;
      register int i; 
      register u_char *cp,*dp;
/*#if 0*/
      register u_long l;
      register int j;
      static int old_rrlen;
      static char old_rr[MAX_IPOPTLEN];
/*#endif*/
      struct iphdr *ip;
      struct timeval tv, *tp;
      long triptime = 0;
      int hlen, dupflag;
	
	ip = (struct iphdr *)buf;
      hlen = ip->ip_hl << 2;
      (void)fprintf(stderr, "testraw : icmp packet (%d bytes) from %s\n", cc,
                    inet_ntoa(*(struct in_addr *)&from->sin_addr.s_addr));

	/* Now the ICMP part */
      cc -= hlen;
      icp = (struct icmphdr *)(buf + hlen);
	if (icp->type == ICMP_ECHOREPLY) {
		(void)fprintf(stderr,"testraw : Got icmp packet with ECHO REPLY \n");
		return;
	}else if (icp->type == ICMP_DEST_UNREACH) {
            switch(icp->code) {
            case ICMP_NET_UNREACH:
                  (void)printf("Destination Net Unreachable\n");
                  break;
            case ICMP_HOST_UNREACH:
                  (void)printf("Destination Host Unreachable\n");
                  break;
            case ICMP_PROT_UNREACH:
                  (void)printf("Destination Protocol Unreachable\n");
                  break;
            case ICMP_PORT_UNREACH:
                  (void)printf("Destination Port Unreachable\n");
                  break;
            case ICMP_FRAG_NEEDED:
                  (void)printf("frag needed and DF set\n");
			printf("MTU : %d \n", icp->un.frag.mtu);
			printf("MTU (ntohs) : %d \n", ntohs(icp->un.frag.mtu));
                  break;
            case ICMP_SR_FAILED:
                  (void)printf("Source Route Failed\n");
                  break;
            case ICMP_NET_UNKNOWN:
                  (void)printf("Network Unknown\n");
                  break;
            case ICMP_HOST_UNKNOWN:
                  (void)printf("Host Unknown\n");
                  break;
            case ICMP_HOST_ISOLATED:
                  (void)printf("Host Isolated\n");
                  break;
            case ICMP_NET_UNR_TOS:
                  printf("Destination Network Unreachable At This TOS\n");
                  break;
            case ICMP_HOST_UNR_TOS:
                  printf("Destination Host Unreachable At This TOS\n");
                  break;
            case ICMP_PKT_FILTERED:
                  (void)printf("Packet Filtered\n");
                  break;
            case ICMP_PREC_VIOLATION:
                  (void)printf("Precedence Violation\n");
                  break;
            case ICMP_PREC_CUTOFF:
                  (void)printf("Precedence Cutoff\n");
                  break;
            default:
                  (void)printf("Dest Unreachable, Unknown Code: %d\n",icp->code);
                  break;
      	}
	}else{ 
		(void)fprintf(stderr,"testraw : Got icmp packet with type %d\n",icp->type);
            (void)fprintf(stderr,"testraw : Unknown Code: %d\n", icp->code);
	}	
	
	return;
}

int
main(int argc, char *argv[])
{
      struct timeval timeout;
      struct hostent *hp;
      struct sockaddr_in *to;
	struct sockaddr_in from;
      struct protoent *proto;
      struct in_addr ifaddr;
      int cc,i, s,is, protoid=0, fromlen;
      int ch, fdmask, hold, packlen, preload;
      u_char *datap, *packet;
      char *target, hnamebuf[MAXHOSTNAMELEN];
      u_char ttl, loop;
      int am_i_root;

	register SCTP_common_header *sctphd;
	register struct iphdr *iphd;

  
//     static char *null = NULL;
//      __environ = &null;
      am_i_root = (getuid()==0);

      /*
       * Pull this stuff up front so we can drop root if desired.
       */
      if (!(proto = getprotobyname("sctp"))) {
            (void)fprintf(stderr, "testraw: unknown protocol sctp - using 132\n");
		protoid = 132;
      }     
	if (!protoid) protoid = proto->p_proto;

      if ((s = socket(AF_INET, SOCK_RAW, protoid)) < 0) {
            if (errno==EPERM) {
                  fprintf(stderr, "testraw: must run as root or setuid root\n");
            }     
            else perror("testraw: socket");
            exit(2);
      }
	/*--------------------------------------------------------------*/
	/* Now also open a socket for capturing ICMP messages		    */
	/*--------------------------------------------------------------*/
	if (!(proto = getprotobyname("icmp"))) {
            (void)fprintf(stderr, "ping: unknown protocol icmp.\n");
            exit(2);
      }

      if ((is = socket(AF_INET, SOCK_RAW, proto->p_proto)) < 0) {
            if (errno==EPERM) {
                  fprintf(stderr, "ping: ping must run as root\n");
            }
            else perror("ping: socket");
            exit(2);
      }
	ch = ICMP_DEST_UNREACH; 
	if (setsockopt(is, SOL_RAW, ICMP_FILTER, (char *)&ch, sizeof(ch)) < 0){
		perror("setsockopt: ICMP_FILTER");
            exit(-1);
      }

	/*--------------------------------------------------------------*/

	/* Now we have a socket -- make packet and send it */ 
   	if (argc != 2){
		printf("ARGC != 2 \n");
            usage();
	}

     	target = (argv[1]);
	printf("Got a target %s\n",target);

      memset(&whereto, 0, sizeof(struct sockaddr));
      to = (struct sockaddr_in *)&whereto;
      to->sin_family = AF_INET;
      if (inet_aton(target, &to->sin_addr)) {
            hostname = target;
		printf("Got a target %s \n",target);
		printf("Got a hostname %s \n",hostname);
      }else 
		usage();

	packlen = MY_PACKETLEN;	// 1500 Bytes FIX
      packet = malloc((u_int)packlen);

      if (!packet) {
            (void)fprintf(stderr, "testraw: out of memory.\n");
            exit(2);
      }     
	
	// Set buffer size to 48 Packets a 1024 Bytes !!!
     hold = 48 * 1024;
     if (setsockopt(s, SOL_SOCKET, SO_RCVBUF, (char *)&hold, sizeof(hold))<0){
	      perror("setsockopt: SO_RCVBUF");
            exit(-1);
      }

	ch = 1; // enable !
	if(setsockopt(s, SOL_IP, IP_HDRINCL, (char *)&ch, sizeof(ch))<0){
		        perror("setsockopt: IP_HDRINCL");
            exit(-1);
	}
	ch = IP_PMTUDISC_DO;
	//if (setsockopt(s, SOL_IP, IP_PMTU_DISCOVER, (char *)&ch, sizeof(ch))<0){
	if (setsockopt(s, SOL_IP, IP_MTU_DISCOVER, (char *)&ch, sizeof(ch))<0){
		perror("setsockopt: IP_PMTU_DISCOVER");
            exit(-1);
      }  
	iphd = (struct iphdr*)outpack;
	iphd->protocol = 132;
	iphd->version = 4;
	iphd->ihl= 5;
	iphd->tos = 0;
	iphd->tot_len = 0;
	iphd->id = 0;
	iphd->ttl = 64;
	iphd->check = 0;
								  
	iphd->daddr = (int)to->sin_addr.s_addr;
	iphd->saddr = ntohl(0xC0A801D7); /* 192.168.1.215 */
	fprintf(stderr, "sizeof(iphdr)= %d\n",sizeof(struct iphdr)); 
	
      sctphd = (SCTP_common_header *)&outpack[sizeof(struct iphdr)];
	
	sctphd->src_port = htons(8080);
	sctphd->dest_port = htons(2905);
      sctphd->checksum = 0;   /* either 0 or valid crc */
      sctphd->verification_tag = htonl(0x12345678);
	(void)gettimeofday((struct timeval *)&outpack[sizeof(struct iphdr)+
					sizeof(SCTP_common_header)],
                			(struct timezone *)NULL);
	cc = sizeof(struct iphdr) + sizeof(SCTP_common_header) + 
		sizeof(struct timeval);
	// 1000 is length of IP data field !! add 20 Bytes for IP Header....
	i = sendto(s, (char *)outpack, cc, 0, &whereto, sizeof(struct sockaddr));

     	if (i < 0 || i != cc)  {
            if (i < 0)
                  perror("testraw: sendto");
	} 
      (void)printf("testraw: wrote %s %d chars, ret=%d\n",hostname, cc, i);
      (void)printf("icmp socket: %d, packlen==%d\n",is,packlen);

/*---------------------------------------------------------------------*/
	fromlen = sizeof(struct sockaddr_in);
	if ((cc = recvfrom(is, (char *)packet, packlen, 0,
                (struct sockaddr *)&from, &fromlen)) < 0) {
                  if (errno == EINTR){
                        fprintf(stderr,"testraw: EINTR in recvfrom \n");
			}
                  perror("testraw: recvfrom/icmp");
			exit(-1);
      }     
	pr_pack((char *)packet, cc, &from);
	exit(0);
} 

/*---------------------------------------------------------------------*/

