#include <sys/param.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <sys/time.h>
#include <sys/signal.h>

#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <arpa/inet.h>

#include <linux/icmpv6.h>


#include <netdb.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <errno.h>

// #include <linux/icmp.h>
#ifndef ICMP_FILTER
	#define ICMP_FILTER                     1
#endif

typedef struct SCTP_COMMON_HEADER
{
     unsigned short src_port;
     unsigned short dest_port;
     unsigned long verification_tag;
     unsigned long checksum;
} SCTP_common_header;


//  #define ICMP_MINLEN     28
#define MY_PACKETLEN 	1500
struct sockaddr_in6 whereto;      /* who to ping */
/*	struct in6_addr
 *   	{
 *       union
 *       {
 *       	uint8_t         u6_addr8[16];
 *      	uint16_t        u6_addr16[8];
 *      	uint32_t        u6_addr32[4];
 *       } in6_u;
 *    #define s6_addr    in6_u.u6_addr8
 *    #define s6_addr16  in6_u.u6_addr16
 *    #define s6_addr32  in6_u.u6_addr32
 *    #define s6_addr64  in6_u.u6_addr64
 *    };
 *    struct sockaddr_in6
 *    {
 *    	__SOCKADDR_COMMON (sin6_);
 *    	uint16_t sin6_port;      
 *   		uint32_t sin6_flowinfo;    
 *    	struct in6_addr sin6_addr;  
 *   	};
 */                                                     

int datalen = MY_PACKETLEN;
int s;                        /* socket file descriptor */
u_char outpack[MY_PACKETLEN];    
static char *hostname;        
static int ident;       /* process id to identify our packets */


static void
usage(void)
{
      (void)fprintf(stderr, "usage: testv6  host-ipv6-address\n");
      exit(2);
}

/*
 * pr_pack --
 * ALL readers of the ICMP socket get a copy of ALL ICMP packets
 * which arrive ('tis only fair).  Take care of that !
 * Taken from ping.c
 */
void
pr_pack(char *buf, int cc, struct sockaddr_in6 *from)
{     
	register struct icmp6hdr *icp;
      register int i; 
      register u_char *cp,*dp;
/*#if 0*/
      register u_long l;
      register int j;
      static int old_rrlen;
      static char old_rr[MAX_IPOPTLEN];
      static char hostname[MY_PACKETLEN];
/*#endif*/
      struct ip6hdr *ip;
      struct timeval tv, *tp;
      long triptime = 0;
      int hlen, dupflag;
	
	icp = (struct icmp6hdr*)buf;
	ip = (struct ip6hdr*)buf;

      inet_ntop(AF_INET6,(struct in6_addr *)&from->sin6_addr.s6_addr,hostname,MY_PACKETLEN);
      (void)fprintf(stderr, "testraw : icmp packet (%d bytes) from %s\n", cc,hostname);

	/* Now the ICMP part */
      cc -= hlen;
	if (ntohs(icp->icmp6_type) == ICMPV6_ECHO_REPLY) {
		(void)fprintf(stderr,"testraw : Got icmpv6 packet with ECHO REPLY \n");
		return;
	}else if (ntohs(icp->icmp6_type) == ICMPV6_DEST_UNREACH) {
            switch(ntohs(icp->icmp6_code)) {
		case ICMPV6_NOROUTE :
                  (void)printf("No route\n");
                  break;
		case ICMPV6_NOT_NEIGHBOUR :
                  (void)printf("Not neighbour\n");
                  break;
            case ICMPV6_ADDR_UNREACH:
                  (void)printf("Address Unreachable\n");
                  break;
            case ICMPV6_PORT_UNREACH:
                  (void)printf("Destination Port Unreachable\n");
                  break;
            default:
                  (void)printf("Dest Unreachable, Unknown Code: %d\n",icp->icmp6_code);
                  break;
      	}
	}else{ 
		(void)fprintf(stderr,"testraw : Got icmp packet with type %d\n",icp->icmp6_type);
            (void)fprintf(stderr,"testraw : Unknown Code: %d\n", icp->icmp6_code);
	}	
	
	return;
}

int
main(int argc, char *argv[])
{
      struct timeval timeout;
      struct hostent *hp;
      struct sockaddr_in6 *to;
	struct sockaddr_in6 from;
      struct protoent *proto;
      struct in6_addr ifaddr;
      struct in6_addr addr;
      int cc,i, s,is, protoid=0, fromlen;
      int ch, fdmask, hold, packlen, preload;
      u_char *datap, *packet;
      char *target, hnamebuf[MAXHOSTNAMELEN];
      u_char ttl, loop;
      int am_i_root;

	register SCTP_common_header *sctphd;
	register struct ip6_hdr *iphd;

  
//     static char *null = NULL;
//      __environ = &null;
      am_i_root = (getuid()==0);

      /*
       * Pull this stuff up front so we can drop root if desired.
       */
      if (!(proto = getprotobyname("sctp"))) {
            (void)fprintf(stderr, "testraw: unknown protocol sctp - using 132\n");
		protoid = 132;
      }     
	if (!protoid) protoid = proto->p_proto;

      if ((s = socket(AF_INET6, SOCK_RAW, protoid)) < 0) {
            if (errno==EPERM) {
                  fprintf(stderr, "testraw: must run as root or setuid root\n");
            }     
            else perror("testv6: raw socket");
            exit(2);
      }
	/*--------------------------------------------------------------*/
	/* Now also open a socket for capturing ICMP messages		    */
	/*--------------------------------------------------------------*/
	if (!(proto = getprotobyname("ipv6-icmp"))) {
            (void)fprintf(stderr, "ping: unknown protocol icmp.\n");
            exit(2);
      }

      if ((is = socket(PF_INET6, SOCK_RAW, proto->p_proto)) < 0) {
            if (errno==EPERM) {
                  fprintf(stderr, "ping: ping must run as root\n");
            }
            else perror("testv6: icmp6 socket");
            exit(2);
      }

	/*--------------------------------------------------------------*/

	/* Now we have a socket -- make packet and send it */ 
   	if (argc != 2){
		printf("ARGC != 2 \n");
            usage();
	}

     	target = (argv[1]);
	printf("Got a target %s\n",target);

      memset(&whereto, 0, sizeof(struct sockaddr_in6));
      to = (struct sockaddr_in6 *)&whereto;
      to->sin6_family = AF_INET6;
	to->sin6_flowinfo = htonl(0x1);
	to->sin6_port = htons(132);
	
	if (inet_pton(AF_INET6, target, &to->sin6_addr) <= 0)
	{

            hp = gethostbyname2(target, AF_INET6);
		if (hp == NULL)
		{
			fprintf(stderr, "unkown host\n");
			exit(1);
		}
	      memcpy(&to->sin6_addr, hp->h_addr_list[0], hp->h_length);
	}else 
		usage();

	packlen = MY_PACKETLEN;	// 1500 Bytes FIX
      packet = malloc((u_int)packlen);

      if (!packet) {
            (void)fprintf(stderr, "testraw: out of memory.\n");
            exit(2);
      }     
	
	// Set buffer size to 48 Packets a 1024 Bytes !!!
     hold = 48 * 1024;
     if (setsockopt(s, SOL_SOCKET, SO_RCVBUF, (char *)&hold, sizeof(hold))<0){
	      perror("setsockopt: SO_RCVBUF");
            exit(-1);
      }

 /*     ch = 12;  an offset for the checksum calculation 
      setsockopt(s, IPPROTO_IPV6, IPV6_CHECKSUM, &ch, sizeof(ch)); */
			   
      sctphd = (SCTP_common_header *)outpack;
	sctphd->src_port = 0x1234;
      sctphd->dest_port = 80;
      sctphd->verification_tag = 0x12345678;
      sctphd->checksum = 0xFF00FF00;
	//	cc = sizeof(struct iphdr) + sizeof(SCTP_common_header) + 
	//	sizeof(struct timeval);
	cc = 1000;
	// 1000 is length of IP data field !! add 20 Bytes for IP Header....
	i = sendto(s, (char *)outpack, cc, 0, &whereto, sizeof(struct sockaddr_in6));

     	if (i < 0 || i != cc)  {
            if (i < 0)
                  perror("testraw: sendto");
	} 
      (void)printf("testraw: wrote %s %d chars, ret=%d\n",hostname, cc, i);
      (void)printf("icmp socket: %d, packlen==%d\n",is,packlen);

/*---------------------------------------------------------------------*/
	fromlen = sizeof(struct sockaddr_in6);
	if ((cc = recvfrom(is, (char *)packet, packlen, 0,
                (struct sockaddr *)&from, &fromlen)) < 0) {
                  if (errno == EINTR){
                        fprintf(stderr,"testraw: EINTR in recvfrom \n");
			}
                  perror("testraw: recvfrom/icmp");
			exit(-1);
      }     
	pr_pack((char *)packet, cc, &from);
	exit(0);
} 

/*---------------------------------------------------------------------*/

